//
//  GameScene.swift
//  SpaceGame
//
//  Created by Paul Solt on 1/26/15.
//  Copyright (c) 2015 Paul Solt. All rights reserved.
//

import SpriteKit

class GameScene: SKScene, SKPhysicsContactDelegate {
    
    // bit mask used for collision detection
    let shipCategory: UInt32 = 0x1 << 0
    let asteroidCategory: UInt32 = 0x1 << 1
    
    var score = 0
    var scoreLabel: SKLabelNode!
    
    // Declare a property to share between methods
    var playerShip: SKSpriteNode!
    
    var gameplayPaused: Bool = false
    
    var asteroidLayer: SKNode!  // Organize all asteroids onto this layer (easy to pause a layer)
    
    var background: SKSpriteNode! // needs to be initialzed before using
    
    // UI
    var pauseButton: UIButton!
    
    
    
    
    override func didChangeSize(oldSize: CGSize) {
     
        // adjust game logic
        background?.size = frame.size  // optional chaining
        
        // Calculate a new position based on half the width and height (position is the center of the image)
        //  so you need to offset the image to display for the full screen
        background?.position = CGPoint(x: frame.size.width / 2.0, y: frame.size.height / 2.0)

    }
    
    
    func reset() {
        
        // start in initial state
        
        resumeGameplay()  // run the game right away
        
        score = 0
        scoreLabel.text = "Score: \(score)"
        
        // remove asteroids?
        // reset player location?
        // animation to show re-spawn or a new game
        
    }
    
    // pause and resume method
    
    func pauseButtonPressed(sender: AnyObject) {
        
        if !gameplayPaused {
            pauseGameplay()
        } else {
            resumeGameplay()
        }
    }
    
    func pauseGameplay() {
        
        physicsWorld.speed = 0 // pause physics
        asteroidLayer.paused = true
        
        gameplayPaused = true
        
    }
    
    func resumeGameplay() {
        
        physicsWorld.speed = 1 // resume physics
        asteroidLayer.paused = false
        
        gameplayPaused = false
    }
    
    
    
    override func didMoveToView(view: SKView) {
        
        
        physicsWorld.contactDelegate = self
        
        // 1. slower gravity force (easier when "smaller" negative number)
        physicsWorld.gravity = CGVector(dx: 0, dy: -1) // dy: -7)
        
        
        // Store the width and height of screen for math calculations
        let width = UIScreen.mainScreen().bounds.size.width   // frame.size.width
        let height = UIScreen.mainScreen().bounds.size.height
        
        // Add the background before the ship, so it is displayed below it
        background = SKSpriteNode(imageNamed: "GalaxyBackground")
        
        // Center the background in the middle of the screen
        background.position = CGPoint(x: width / 2.0, y: height / 2.0)
        
        // Make the background the same size of the screen (stretch and centers)
        background.size = CGSize(width: width, height: height)
        
        background.zPosition = 0
        
        addChild(background)
        
        
        // Load ship image and display it
        playerShip = SKSpriteNode(imageNamed: "PlayerShip1")
        playerShip.position = CGPoint(x: 200, y: 400)
        
        playerShip.physicsBody = SKPhysicsBody(texture: playerShip.texture, size: playerShip.size)
        
        // 2. Changing physics for the player ship can change gameplay
        playerShip.physicsBody?.dynamic = false // can have collisions if true, otherwise it's immovable to asteroids
        playerShip.physicsBody?.affectedByGravity = false // not affected by gravity when false
        
        // collision rules
        playerShip.physicsBody?.categoryBitMask = shipCategory  // ship
        playerShip.physicsBody?.collisionBitMask = asteroidCategory
        playerShip.physicsBody?.contactTestBitMask = asteroidCategory
        playerShip.zPosition = 1
        
        
        // Create an animation using SKAction that can loop (repeat)
        var colorAction1 = SKAction.colorizeWithColor(SKColor.redColor(), colorBlendFactor: 0.75, duration: 2)
        var colorAction2 = SKAction.colorizeWithColor(SKColor.whiteColor(), colorBlendFactor: 0.0, duration: 2)
        var colorActionSequence = SKAction.sequence([colorAction1, colorAction2])
        var colorAnimation = SKAction.repeatActionForever(colorActionSequence)
        
        playerShip.runAction(colorAnimation, withKey: "ColorAnimation")
        
        
        addChild(playerShip)
        
        
        // 3. asteroid creation actions (spawning)
        
        asteroidLayer = SKNode()
        asteroidLayer.zPosition = 2
        addChild(asteroidLayer)
        
        // run code, wait, repeat sequence (spawning asteroids)
        
        // create an asteroid and add it to the scene (block of code to execute during time interval)
        let asteroidCreateAction = SKAction.runBlock { () -> Void in
            let asteroid = self.createAsteroid()
            self.asteroidLayer.addChild(asteroid)
        }
        
        // Create a way to control the asteroid spawn speed using a user friendly value (i.e. 2 per second)
        let asteroidsPerSecond: CGFloat = 2
        let asteroidInterval: NSTimeInterval = NSTimeInterval(1.0 / asteroidsPerSecond)
        let asteroidWaitAction = SKAction.waitForDuration(asteroidInterval, withRange: 0.0)
        
        // Using a sequence and a repeat can make it continue forever
        let asteroidSequenceAction = SKAction.sequence([asteroidCreateAction, asteroidWaitAction])
        let asteroidRepeatAction = SKAction.repeatActionForever(asteroidSequenceAction)
        
        asteroidLayer.runAction(asteroidRepeatAction)
        
        scoreLabel = SKLabelNode(text: "Score: \(score)")
        
        scoreLabel.position = CGPoint(x: frame.width / 2.0, y: frame.height - scoreLabel.calculateAccumulatedFrame().height - 20)
        scoreLabel.zPosition = 10 // top
        
        score = 0 // reset method?
        
        addChild(scoreLabel)
    }
    
    override func touchesBegan(touches: NSSet, withEvent event: UIEvent) {
        
        if !gameplayPaused {
            
            if let touch = touches.anyObject() as UITouch? {
                let location = touch.locationInNode(self)
                //            let move = SKAction.moveTo(location, duration: 0.25)
                
                let distance = distanceBetweenPoints(playerShip.position, b: location)
                
                println("distance: \(distance)")
                
                var playerSpeed: CGFloat = 700  // points per second //CGFloat(300)
                
                let time = timeToTravelDistance(distance, speed: playerSpeed)
                
                let move = SKAction.moveTo(location, duration: time)
                //move.timingMode = SKActionTimingMode.EaseInEaseOut
                println("time: \(time)")
                
                // need an instance variable (property) to move the ship
                playerShip.runAction(move)
            }
            
        }
    }
    
    // Player movement change (drag to move instead of tap) - i.e. bomb dodge style: https://itunes.apple.com/us/app/protect-the-bomb/id660318608?mt=8
    //    override func touchesMoved(touches: NSSet, withEvent event: UIEvent) {
    //
    //        // drag ship to location
    //        if let touch = touches.anyObject() as UITouch? {
    //            let location = touch.locationInNode(self)
    //
    //            playerShip.position = location
    //        }
    //    }
    
    
    
    // c = square root(a^2 + b^2)
    
    // Calculate the distance betwen two points
    // @return linear distance between two CGPoints
    func distanceBetweenPoints(a: CGPoint, b: CGPoint) -> CGFloat {
        return sqrt((b.x - a.x)*(b.x - a.x) + (b.y - a.y)*(b.y - a.y))
    }
    
    // calculate time using the rate * time = distance formula (solve for time)
    // return a NSTimeInterval to signify time (in seconds as a return value)
    func timeToTravelDistance(distance: CGFloat, speed: CGFloat) -> NSTimeInterval {
        let time = distance / speed
        return NSTimeInterval(time) // NSTimeInterval(distance / speed)
    }
    
    
    func createAsteroid() -> SKSpriteNode {
        
        // Make the asteroids interesting with random sizes (scale), angular velocities, different images,
        //  color effects (color + colorBlendFactor), x velocity, y velocity
        
        
        // Use random numbers
        //  a. Random integers use: arc4random_uniform(10)  // makes the range [0, 9]
        /// b. Random floating point values: drand48() // makes range [0, 1]
        //      you'll need to seed drand48() with srand48() using a random/changing number (i.e. time)
        
        // Try printing out values to get an idea of what they look like
        //        println("random: \(arc4random_uniform(10))")
        
        // When working with random numbers in Swift, you will have to do a lot of type conversions
        //  Different functions use different types, most types in Sprite Kit are CGFloat, so
        //  you will need to "massage" the types into CGFloat to use in-game.
        
        
        // 1. Use an array and choose between two (or more) different image files from asset catalog
        var images = ["Asteroid1",
            "Asteroid2"]
        var index: Int = Int(arc4random_uniform(UInt32(images.count)))
        var asteroidName = images[index]  // use the index to get an image name from the array
        
        var asteroid = SKSpriteNode(imageNamed: asteroidName)
        
        // Set a depth value for the asteroids, so they are on top of the background
        //  but below UI/player. Bigger positive numbers are on the top
        asteroid.zPosition = 2
        
        // 2. Create a random starting x position
        
        var randomX = CGFloat(arc4random_uniform(UInt32(frame.size.width)))  // 1.0, 1
        asteroid.position.x = randomX
        asteroid.position.y = frame.size.height + asteroid.size.height  // move asteroid to top and offscreen
        
        //        println("width: \(frame.size)")
        // 3. Change the scale (use same for both xScale and yScale to prevent stretching/squashing
        //    Create a range and limit the scale
        //      drand48() // range is [0.0, 1.0]
        //      drand48() * 0.5 // range is [0.0, 0.5]
        //      drand48() * 0.5 + 0.2 // range is [0.2, 0.7]
        // min() or max() can help you limit numbers by either taking the smaller or larger number of two values
        //      min(CGFloat(drand48() * 0.5) + 0.2, 1) // using min([0.2, 0.7], 1)
        //      the value is always less than 1, so the range is just [0.2, 0.7]
        asteroid.xScale = min(CGFloat(drand48() * 0.5) + 0.2, 1)
        asteroid.yScale = asteroid.xScale
        //        println("Scale: \(asteroid.xScale)") // print the values to see what kind of sizes you're creating
        
        // 4. Try random color (a little crazy on the eyes, maybe not as wonky colors)
        //        asteroid.colorBlendFactor = 0.5
        //        asteroid.color = SKColor(red: CGFloat(drand48()), green: CGFloat(drand48()), blue: CGFloat(drand48()), alpha: 1.0)
        
        // 5. Create a random angular velocity (i.e. spin the asteroid so it doesn't look strange)
        //    Use a random float [-1.0, 1.0] * maxSpeed = +/- [0, maxSpeed]
        var maxAngularVelocity: CGFloat = CGFloat(2.0 * M_PI) // 2PI = 1 revolution/second
        var angularVelocity: CGFloat = CGFloat(drand48() * 2.0 - 1.0) * maxAngularVelocity
        //        println("Angle: \(angularVelocity)") // print out the angular velocity to check values
        
        // 6. Change the x or y speeds using random values, feel free to tweak these values until you find something
        //      that feels good to you. Play testing requires a lot of tweaking to get the right numbers
        //      otherwise it'll be too hard or too easy
        
        var maxXSpeed: CGFloat = 10
        var xSpeed: CGFloat = CGFloat(drand48() * 2.0 - 1.0) * maxXSpeed
        var maxYSpeed: CGFloat = -200
        var ySpeed: CGFloat = CGFloat(drand48()) * maxYSpeed
        
        // add physics after making any changes to the sprite size/image
        
        // add after any size or image changes
        asteroid.physicsBody = SKPhysicsBody(texture: asteroid.texture, size: asteroid.size)
        
        asteroid.physicsBody?.categoryBitMask = asteroidCategory
        asteroid.physicsBody?.collisionBitMask = shipCategory | asteroidCategory // use all interactions between asteroids + ships
        asteroid.physicsBody?.contactTestBitMask = shipCategory //| asteroidCategory  // You might want asteroid to break apart on collision
        
        // 7.0 Set the angular velocity and velocity using the random values from above (must be on a valid physics body)
        //  you'll set it here after you create the physics body.
        asteroid.physicsBody?.angularVelocity = (angularVelocity)
        asteroid.physicsBody?.velocity = CGVector(dx: xSpeed, dy: ySpeed)
        
        // 8.0 - set the name, so that you can search for the object in the game scene to remove it
        asteroid.name = "asteroid"
        
        return asteroid
    }
    
    
    override func update(currentTime: NSTimeInterval) {
        //        println("time: \(currentTime)")
        
    }
    
    override func didSimulatePhysics() {
        
        // is an asteroid offscreen? remove it using the enumerateChildNodes method
        
        asteroidLayer.enumerateChildNodesWithName("asteroid", usingBlock: { (asteroid: SKNode!, stop: UnsafeMutablePointer<ObjCBool>) -> Void in
            
            //            println("found an asteroid")
            
            // Only remove after they pass below the edge of the screen (height/2 would be more exact, but this gives some le-way to prevent graphics glitches)
            if asteroid.position.y < (0 - asteroid.frame.height) {
                
                // offscreen
                asteroid.removeFromParent()
                
                // Game rule: Add a point for each asteroid that was passed
                self.score = self.score + 1
                self.scoreLabel.text = "Score: \(self.score)"
            }
            
        })
    }
    
    func didBeginContact(contact: SKPhysicsContact) {
        // a collision happened
        //        println("collision")  // print when it happens as a sanity check
        
        // Collision response for an asteroid touching the player's ship
        if contact.bodyA.categoryBitMask == shipCategory && contact.bodyB.categoryBitMask == asteroidCategory ||
            contact.bodyA.categoryBitMask == asteroidCategory && contact.bodyB.categoryBitMask == shipCategory {
                
                // gameover condition or life lost condition
                
                //  it's up to you how to handle it. You might want to trigger a player respawn using action sequences
                //  e.g. action sequence pseudocode: [remove ship, wait, add ship, flash ship alpha/colorize it, begin game]
                
                // Update score (reset here or later on if there are multiple lives)
                score = 0 // reset score and notify player
                scoreLabel.text = "Score: \(score)"
                
                // Optional: You could remove the asteroid from the game when it touches
                
                // Body a is asteroid
                //                if contact.bodyA.categoryBitMask == asteroidCategory {
                //                    contact.bodyA.node?.removeFromParent()
                //                }
                // Body b is asteroid
                //                if contact.bodyB.categoryBitMask == asteroidCategory {
                //                    contact.bodyB.node?.removeFromParent()
                //                }
        }
        
        
        // Optional physics response code to change behavior of game or points
        if contact.bodyA.categoryBitMask == asteroidCategory && contact.bodyB.categoryBitMask == asteroidCategory ||
            contact.bodyA.categoryBitMask == asteroidCategory && contact.bodyB.categoryBitMask == asteroidCategory {
                
                // You could do something with the force, direction, and point of cnotact between two
                // asteroids. It's up to you how to handle collisions
                
                var force = contact.collisionImpulse
                var direction = contact.contactNormal
                var point = contact.contactPoint
                
                // e.g.: create two new asteroids that are smaller for each big asteroid
        }
    }
    
    func didEndContact(contact: SKPhysicsContact) {
        
    }
    
}
